/*
Copyright 2003-2009 Aris Adamantiadis

This file is part of the SSH Library

You are free to copy this file, modify it in any way, consider it being public
domain. This does not apply to the rest of the library though, but it is
allowed to cut-and-paste working code from this file to any license of
program.
The goal is to show the API in action. It's not a reference on how terminal
clients must be made or how a client should react.
*/

#include "config.h"

//#include <sys/statvfs.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#else
/*
 * Declared in <unistd.h>
 */
char *optarg;       /* argument associated with option */
int opterr = 1;     /* if error message should be printed */
int optind = 1;		/* index into parent argv vector */
int optopt;         /* character checked for validity */

int optreset;       /* reset getopt */

#define	BADCH	(int)'?'
#define	BADARG	(int)':'
#define	EMSG	""

/*******************************************************************************
* wceex_getopt - function is a command-line parser
*
* Description:
*
*   The parameters argc and argv are the argument count and argument array as
*   passed to main(). The argument optstring is a string of recognised option
*   characters. If a character is followed by a colon, the option takes an argument.
*
*   The variable optind is the index of the next element of the argv[] vector to be
*   processed. It is initialised to 1 by the system, and getopt() updates it when it
*   finishes with each element of argv[].
*
*   NOTE: Implementation of the getopt() function was grabbed from NetBSD
*         operating system sources. See copyright note in the file header.
*
* Return:
*
*   Returns the next option character specified on the command line.
*   A colon (:) is returned if getopt() detects a missing argument and
*   the first character of optstring was a colon (:).
*   A question mark (?) is returned if getopt() encounters an option
*   character not in optstring or detects a missing argument and the first
*   character of optstring was not a colon (:).
*   Otherwise getopt() returns -1 when all command line options are parsed. 
* 
* Reference:
*
*   IEEE 1003.1, 2004 Edition
*
*******************************************************************************/

static int getopt(int argc, char * const argv[], const char *optstring)
{
	static char *place = EMSG;		/* option letter processing */
	char *oli;				/* option letter list index */

	if (optreset || *place == 0) {		/* update scanning pointer */
		optreset = 0;
		place = argv[optind];
		if (optind >= argc || *place++ != '-') {
			/* Argument is absent or is not an option */
			place = EMSG;
			return (-1);
		}
		optopt = *place++;
		if (optopt == '-' && *place == 0) {
			/* "--" => end of options */
			++optind;
			place = EMSG;
			return (-1);
		}
		if (optopt == 0) {
			/* Solitary '-', treat as a '-' option
			   if the program (eg su) is looking for it. */
			place = EMSG;
			if (strchr(optstring, '-') == NULL)
				return -1;
			optopt = '-';
		}
	} else
		optopt = *place++;

	/* See if option letter is one the caller wanted... */
	if (optopt == ':' || (oli = strchr(optstring, optopt)) == NULL) {
		if (*place == 0)
			++optind;
		if (opterr && *optstring != ':')
			(void)fprintf(stderr,
                                      "unknown option -- %c\n", optopt);
		return (BADCH);
	}

	/* Does this option need an argument? */
	if (oli[1] != ':') {
		/* don't need argument */
		optarg = NULL;
		if (*place == 0)
			++optind;
	} else {
		/* Option-argument is either the rest of this argument or the
		   entire next argument. */
		if (*place)
			optarg = place;
		else if (argc > ++optind)
			optarg = argv[optind];
		else {
			/* option-argument absent */
			place = EMSG;
			if (*optstring == ':')
				return (BADARG);
			if (opterr)
				(void)fprintf(stderr,
                                        "option requires an argument -- %c\n",
                                        optopt);
			return (BADCH);
		}
		place = EMSG;
		++optind;
	}
	return (optopt);			/* return option letter */
}


#endif


#include <libssh/libssh.h>
#include <libssh/sftp.h>
#include <libssh/callbacks.h>
#include <libssh/misc.h>


#include "examples_common.h"
#ifdef WITH_SFTP

static int verbosity;
static char *destination;

#ifdef UNDER_CE
int errno = 0;
#endif

static void WinCeLoggingCallback(int verbosity, const char *function, const char *buffer, void *userdata)
{
    printf("[%d] %s:  %s\n", verbosity, function, buffer);
}

#define DATALEN 65536
static void do_sftp(ssh_session session){
    sftp_session sftp=sftp_new(session);
    sftp_dir dir;
    sftp_attributes file;
    sftp_statvfs_t sftpstatvfs;
    //struct statvfs sysstatvfs;
    sftp_file fichier;
    sftp_file to;
    int len=1;
    unsigned int i;
    static char data[DATALEN]={0};
    char *lnk;

    unsigned int count;

    if(!sftp){
        fprintf(stderr, "sftp error initialising channel: %s\n",
            ssh_get_error(session));
        return;
    }
    if(sftp_init(sftp)){
        fprintf(stderr, "error initialising sftp: %s\n",
            ssh_get_error(session));
        return;
    }

    printf("Additional SFTP extensions provided by the server:\n");
    count = sftp_extensions_get_count(sftp);
    for (i = 0; i < count; i++) {
      printf("\t%s, version: %s\n",
          sftp_extensions_get_name(sftp, i),
          sftp_extensions_get_data(sftp, i));
    }

    /* test symlink and readlink */
    if (sftp_symlink(sftp, "/tmp/this_is_the_link",
          "/tmp/sftp_symlink_test") < 0) {
      fprintf(stderr, "Could not create link (%s)\n", ssh_get_error(session));
      return;
    }

    lnk = sftp_readlink(sftp, "/tmp/sftp_symlink_test");
    if (lnk == NULL) {
      fprintf(stderr, "Could not read link (%s)\n", ssh_get_error(session));
      return;
    }
    printf("readlink /tmp/sftp_symlink_test: %s\n", lnk);

    sftp_unlink(sftp, "/tmp/sftp_symlink_test");

    if (sftp_extension_supported(sftp, "statvfs@openssh.com", "2")) {
      sftpstatvfs = sftp_statvfs(sftp, "/tmp");
      if (sftpstatvfs == NULL) {
        fprintf(stderr, "statvfs failed (%s)\n", ssh_get_error(session));
        return;
      }

      printf("sftp statvfs:\n"
          "\tfile system block size: %llu\n"
          "\tfundamental fs block size: %llu\n"
          "\tnumber of blocks (unit f_frsize): %llu\n"
          "\tfree blocks in file system: %llu\n"
          "\tfree blocks for non-root: %llu\n"
          "\ttotal file inodes: %llu\n"
          "\tfree file inodes: %llu\n"
          "\tfree file inodes for to non-root: %llu\n"
          "\tfile system id: %llu\n"
          "\tbit mask of f_flag values: %llu\n"
          "\tmaximum filename length: %llu\n",
          (unsigned long long) sftpstatvfs->f_bsize,
          (unsigned long long) sftpstatvfs->f_frsize,
          (unsigned long long) sftpstatvfs->f_blocks,
          (unsigned long long) sftpstatvfs->f_bfree,
          (unsigned long long) sftpstatvfs->f_bavail,
          (unsigned long long) sftpstatvfs->f_files,
          (unsigned long long) sftpstatvfs->f_ffree,
          (unsigned long long) sftpstatvfs->f_favail,
          (unsigned long long) sftpstatvfs->f_fsid,
          (unsigned long long) sftpstatvfs->f_flag,
          (unsigned long long) sftpstatvfs->f_namemax);

      sftp_statvfs_free(sftpstatvfs);

      //if (statvfs("/tmp", &sysstatvfs) < 0) {
      //  fprintf(stderr, "statvfs failed (%s)\n", strerror(errno));
      //  return;
      //}

      //printf("sys statvfs:\n"
      //    "\tfile system block size: %llu\n"
      //    "\tfundamental fs block size: %llu\n"
      //    "\tnumber of blocks (unit f_frsize): %llu\n"
      //    "\tfree blocks in file system: %llu\n"
      //    "\tfree blocks for non-root: %llu\n"
      //    "\ttotal file inodes: %llu\n"
      //    "\tfree file inodes: %llu\n"
      //    "\tfree file inodes for to non-root: %llu\n"
      //    "\tfile system id: %llu\n"
      //    "\tbit mask of f_flag values: %llu\n"
      //    "\tmaximum filename length: %llu\n",
      //    (unsigned long long) sysstatvfs.f_bsize,
      //    (unsigned long long) sysstatvfs.f_frsize,
      //    (unsigned long long) sysstatvfs.f_blocks,
      //    (unsigned long long) sysstatvfs.f_bfree,
      //    (unsigned long long) sysstatvfs.f_bavail,
      //    (unsigned long long) sysstatvfs.f_files,
      //    (unsigned long long) sysstatvfs.f_ffree,
      //    (unsigned long long) sysstatvfs.f_favail,
      //    (unsigned long long) sysstatvfs.f_fsid,
      //    (unsigned long long) sysstatvfs.f_flag,
      //    (unsigned long long) sysstatvfs.f_namemax);
    }

    /* the connection is made */
    /* opening a directory */
    dir=sftp_opendir(sftp,"./");
    if(!dir) {
        fprintf(stderr, "Directory not opened(%s)\n", ssh_get_error(session));
        return ;
    }
    /* reading the whole directory, file by file */
    while((file=sftp_readdir(sftp,dir))){
        fprintf(stderr, "%30s(%.8o) : %s(%.5d) %s(%.5d) : %.10llu bytes\n",
            file->name,
            file->permissions,
            file->owner,
            file->uid,
            file->group,
            file->gid,
            (long long unsigned int) file->size);
        sftp_attributes_free(file);
    }
    /* when file=NULL, an error has occured OR the directory listing is end of file */
    if(!sftp_dir_eof(dir)){
        fprintf(stderr, "Error: %s\n", ssh_get_error(session));
        return;
    }
    if(sftp_closedir(dir)){
        fprintf(stderr, "Error: %s\n", ssh_get_error(session));
        return;
    }
    /* this will open a file and copy it into your /home directory */
    /* the small buffer size was intended to stress the library. of course, you can use a buffer till 20kbytes without problem */

    fichier=sftp_open(sftp,"/usr/bin/ssh",O_RDONLY, 0);
    if(!fichier){
        fprintf(stderr, "Error opening /usr/bin/ssh: %s\n",
            ssh_get_error(session));
        return;
    }
    /* open a file for writing... */
    to=sftp_open(sftp,"ssh-copy",O_WRONLY | O_CREAT, 0700);
    if(!to){
        fprintf(stderr, "Error opening ssh-copy for writing: %s\n",
            ssh_get_error(session));
        return;
    }
    while((len=sftp_read(fichier,data,4096)) > 0){
        if(sftp_write(to,data,len)!=len){
            fprintf(stderr, "Error writing %d bytes: %s\n",
                len, ssh_get_error(session));
            return;
        }
    }
    printf("finished\n");
    if(len<0)
        fprintf(stderr, "Error reading file: %s\n", ssh_get_error(session));
    sftp_close(fichier);
    sftp_close(to);
    printf("fichiers ferm\n");
    to=sftp_open(sftp,"/tmp/grosfichier",O_WRONLY|O_CREAT, 0644);
    for(i=0;i<1000;++i){
        len=sftp_write(to,data,DATALEN);
        //printf("wrote %d bytes\n",len);
        if(len != DATALEN){
            printf("chunk %d : %d (%s)\n",i,len,ssh_get_error(session));
        }
    }
    sftp_close(to);

    /* close the sftp session */
    sftp_free(sftp);
    printf("sftp session terminated\n");
}

static void usage(const char *argv0){
  fprintf(stderr,"Usage : %s [-v] remotehost\n"
      "sample sftp test client - libssh-%s\n"
      "Options :\n"
      "  -v : increase log verbosity\n",
      argv0,
      ssh_version(0));
  exit(0);
}

static int opts(int argc, char **argv){
  int i;
  while((i=getopt(argc,argv,"v"))!=-1){
    switch(i){
      case 'v':
        verbosity++;
        break;
      default:
        fprintf(stderr,"unknown option %c\n",optopt);
        usage(argv[0]);
        return -1;
    }
  }

  destination=argv[optind];
  if(destination == NULL){
    usage(argv[0]);
    return -1;
  }
  return 0;
}

int main(int argc, char **argv){
  ssh_session session;

  ssh_set_log_callback(WinCeLoggingCallback);

  if(opts(argc,argv)<0)
    return EXIT_FAILURE;
  session=connect_ssh(destination,NULL,verbosity);
  if(session == NULL)
  	return EXIT_FAILURE;
  do_sftp(session);
  ssh_disconnect(session);
  ssh_free(session);
  ssh_finalize();
  return 0;
}



#endif
